/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * Copyright (C) 1997  Tim Edwards
 * Copyright (C) 1999  Sergiy Kovtun
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/*----------------------------------------------------------------*/
/* a821.c							  */
/*----------------------------------------------------------------*/

#include <linux/autoconf.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/types.h> 
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include "a821.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

#define BYTE unsigned char
#define LONG unsigned long

static int adc_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int adc_write(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int adc_open(struct inode *iNode, struct file *filePtr);
static void adc_close(struct inode *iNode, struct file *filePtr);
static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);

int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

/*#define DEBUG 0*/  

#define READY_mask  0x010

 
/* Default device major number */

#ifndef DEFAULT_MAJOR_DEV
#define DEFAULT_MAJOR_DEV  39    /* Default Major Device Number */
#endif

/* Base I/O Address */

#ifndef BASE_REG
#define BASE_REG         0x230   /* Default base address for a821 */
#endif

/* Number of channels */

#ifndef CHANNELS
#define CHANNELS	10
#endif

#define A_MASK		0xff
#define B_MASK		0xff00


typedef struct CHANNEL_REC {
   int           open;        /* for lockout of other processes */
   unsigned char mode;        /* for readback of a821 mode */
   int           busy;        /* for lockout of other minors */  
} ChanRec;

static ChanRec Chan[CHANNELS];              /* Channel specific information   */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in       */
static int portbase;
static int delay[4] = {23,28,140,1300}; /* Gain setting time before read AC data*/
static int lastgain = 0;
/* char kernel_version[] = UTS_RELEASE;  */         /* Required by module             */

#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif


/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int adc_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{ 
  int    i;
  long  value = 0L;
  int   ready = 1;
  int   timeout = 25;
  int   minor = MINOR(iNode->i_rdev);
  portbase = BASE_REG ;  

  switch (minor)
  {
    case 0:
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
      if (Chan[minor].busy == TRUE) printk("Busy");  
      while ( Chan[minor].busy == TRUE)
          { 
	   printk("ADC (%d) is busy\n",minor);   
	   udelay( minor * 5 );
	  }
      for (i =0; i < CHANNELS; i++) Chan[i].busy = TRUE;  
      outb_p( (BYTE)(minor & 0x07), portbase + SET_CH); 
      outb_p( (BYTE)(Chan[minor].mode), portbase + SET_GAIN);
      if (Chan[minor].mode != lastgain) {        /* Gain was changed? */
         lastgain = Chan[minor].mode;   
         udelay(delay[Chan[minor].mode]);   
         }
      /*  Start AD conversion, read a hight byte and check READY-bit */
      outb_p( (BYTE)(0x01), portbase + SOFT_TRIG);
      while (ready != 0) {
        ready = (inb(portbase+AD_HI) & READY_mask);
	udelay( minor );
	timeout--;
	if(timeout == 0) {
           printk("adc_read: Timeout READY in minor number (%d).\n", minor);
           for (i=0; i < CHANNELS; i++) Chan[i].busy = FALSE;   
	   return -1;
	   break;
          }
      }
      
      value = inb(portbase + AD_LO);
      value += (inb(portbase + AD_HI) & 0x0f) * 256;  /* only 12 bits */
      
      for (i=0; i < CHANNELS; i++) Chan[i].busy = FALSE;
      cli();
      put_fs_long(value, buf);
      sti();
      return count;
      break;
   case 9:                         /* Digital input  */
        value  = inb(portbase + DIGITIN_LO);
        value += inb(portbase + DIGITIN_HI) * 256;
        cli();
        put_fs_long(value, buf);
        sti();  
#ifdef DEBUG
        printk("Digital_In get :%#X\n", value & 0xffff);
#endif
                return count;
        break;  
    case 8:                              /* DAC : write only , Error! */
       printk("adc_read: Minor number(%d) :is not readable!\n", minor);
       return -1;
       break;    
    default: 
       printk("adc_read: Incorrect minor number (%d).\n", minor);
       return -1;
       break;
    }
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int adc_write(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
    
    int  minor = MINOR(iNode->i_rdev);
    int  portbase = BASE_REG;        
    long value;
        
    if ( minor > CHANNELS ) {
        printk("adc_write: bad minor number = %d\n", minor );
        return -1;
    }
    
    switch (minor) {
     case 8: 
       cli();
       value = get_fs_long(buf) & 0xfff;
       outb_p( (BYTE) (value & A_MASK),        portbase + DA_CH0_LO);
       outb_p( (BYTE)((value & B_MASK) >> 8),  portbase + DA_CH0_HI); 
       sti();
       
#ifdef DEBUG
           printk("DAC set to :%#X\n", value & 0xffff);
#endif
       return 2;
       break;

     case 9:
       cli();
       value = get_fs_long(buf) & 0xffff;
       outb_p( (BYTE) (value & A_MASK),        portbase + DIGITOUT_LO);
       outb_p( (BYTE)((value & B_MASK) >> 8),  portbase + DIGITOUT_HI); 
       sti();
#ifdef DEBUG
           printk("Digital_Out put :%#X\n", value & 0xffff);
#endif
       return 2;
       break;
 
     default:
       printk("adc_read: Minor number(%d) :is not writeable!\n", minor);
       return(-EINVAL);
       break;
  }
    
}

/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int adc_open(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);
    int portbase = BASE_REG ;  

    /* 
       check if device is already open: only one process may read from a
       port at a time.  There is still the possibility of two processes 
       reading from two different channels messing things up. However,
       the overhead to check for this may not be worth it.
    */

    if ( Chan[minor].open == TRUE ) {
        return -EBUSY;
    }

    Chan[minor].open = TRUE;              /* The device is open */
    Chan[minor].mode = A821_BI_1;
    outb_p(A821_BI_1, portbase + SET_GAIN);

#ifdef DEBUG
     printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor,
                                           Chan[minor].mode);
#endif
    return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void adc_close(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);

#ifdef DEBUG
        printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
#endif

    Chan[minor].open = FALSE;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int portbase = BASE_REG;              

  switch (cmd) {
    case ADC_SET_GAIN:
      Chan[minor].mode = (BYTE)arg;
      outb_p((BYTE)(MODE_SET & arg), portbase + SET_GAIN);
      lastgain = arg; 
#ifdef DEBUG
           printk("minor:(%d) set gain : %d\n", minor, arg & 0xff);
#endif

      break;

    case ADC_GET_GAIN:
      put_fs_long(Chan[minor].mode, (long*) arg);
      break;

    default:
      return(-EINVAL);
      break;
  }
  return 0;
}

/***************************************************************************
 *
 * See /usr/include/linux/fs.h  for more information
 *
 ***************************************************************************/

static struct file_operations adc_fops = {
	NULL,           /* lseek */
	adc_read,	/* read */
	adc_write,      /* write */
	NULL,		/* readdir */
	NULL,		/* select */
	adc_ioctl,	/* ioctl */
	NULL,		/* mmap */
	adc_open,	/* open */
	adc_close,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL		/* revalidate */
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod a821.o" is invoked on the command line.
 *
 ***************************************************************************/

int init_module(void) 
{
  int i, ready ;
  portbase = BASE_REG;
  
        
  /* Register as a device with kernel.  */

  if (register_chrdev(MajorNumber, "adc", &adc_fops)) {
    printk("%s: Failure to load module.\n", ADAPTER_ID);
    return -EIO;
  } 
  
  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < CHANNELS; i++)
    {
    Chan[i].open = FALSE;
    Chan[i].busy = FALSE;
    }
  
  /* A821 card detect  */
 
  outb_p(0x00, portbase + SOFT_TRIG);
  udelay(2);
  outb_p(0x01, portbase + SET_MODE);
  outb_p(0x00, portbase + SOFT_TRIG);
  printk("Detect A821PGH card on address %#x ..........", portbase);
  udelay(1000);
  ready = (inb( portbase + AD_HI ) & READY_mask);
  if (ready == 0) {
         printk(" Done.\n");
	 printk("%s: module installed on %xh\n", ADAPTER_ID, BASE_REG);
      }
    else {
      printk(" Failed.\n");
      i =  (unregister_chrdev(MajorNumber, "adc")) ;
      printk("%s: module not installed on %x\n", ADAPTER_ID, BASE_REG);
      return -EIO;
      }
   return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod a821" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
    if (MOD_IN_USE) {
      printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    }

    if (unregister_chrdev(MajorNumber, "adc") != 0) {
      printk("%s: cleanup_module failed\n", ADAPTER_ID);
     }
     else {
      printk("%s: module removed.\n", ADAPTER_ID);
     }    
}

#ifdef __cplusplus
}
#endif

